/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file applications/afs/database/lookup.c
 * @author Christian Grothoff
 *
 * This module is responsible for locating available content.
 * After a query is received, the bloomfilter is considered
 * to test if the content is likely to be available. If the
 * bloomfilter-test succeeds, the lookup module is invoked
 * to test if the content is there, and if, tell where it
 * is stored. For 3HASH-queries, the lookup module also 
 * produces the double-hash (inverts the hash).
 **/

#include "config.h"
#include <stdio.h>
#include <dirent.h>
#include <errno.h>
#include "util/checksum.h"
#include "util/cron.h"
#include "util/configuration.h"
#include "util/logging.h"
#include "util/returnvalues.h"
#include "util/statuscalls.h"
#include "applications/afs/database/lookup.h"

/**
 * Handle of the top-level hashtable.
 **/
static CollisionLevel * rootLevel;

/**
 * Cron-job that makes sure the lookup-module is fine
 * by synchronizing the in-memory database with the drive.
 **/
static void cronSyncLookup(CollisionLevel * level) {
  LOG(LOG_CRON,
      "CRON: enter cronSyncLookup\n");
#if HAVE_FDATASYNC
  if (fdatasync(level->handle) != 0)
    LOG(LOG_WARNING, 
	"WARNING: fdatasync failed on lookup database (%s)!\n",
	strerror(errno));
#else
  if (fsync(level->handle) != 0)
    LOG(LOG_WARNING,
	"WARNING: fsync failed on database (%s)!\n",
	strerror(errno));
#endif
  LOG(LOG_CRON, 
      "CRON: exit cronSyncLookup\n");
}

/**
 * Initialize the lookup-module.
 **/
void initLookup() {  
  CollisionLevel * tmp;
  char * fileName;
  char * fileNameDotted;
  int i;
  
  fileName = getFileName("AFS",
			 "DATABASE",
			 "Configuration file must specify filename for"\
			 " the database in section %s under %s.\n");
  fileNameDotted = xmalloc(strlen(fileName)+3,
			   "initLookup: database.X filename");
  rootLevel = NULL;
  for (i=LOOKUP_LEVELS-1;i>=0;i--) {
    tmp = xmalloc(sizeof(CollisionLevel),
		  "initLookup: collisionLevel");
    tmp->next = rootLevel; 
    tmp->mask = (CACHE_HASH_TABLE_SIZE >> (i*LOOKUP_POWER_FACTOR))-1;
    tmp->offset = i;
    MUTEX_CREATE(&tmp->lock);
    rootLevel = tmp;
    strcpy(fileNameDotted, 
	   fileName);
    strcat(fileNameDotted, ".0");
    fileNameDotted[strlen(fileNameDotted)-1] += i;
    tmp->handle = open(fileNameDotted, 
		       O_CREAT|O_RDWR, 
		       S_IRUSR|S_IWUSR);  
    if (tmp->handle < 0)
      errexit("Could not open/create database %s (%s)!\n",
	      fileNameDotted,
	      strerror(errno));
    registerSpaceMonitor(fileNameDotted);
  }
  xfree(fileNameDotted,
	"initLookup: dotted filename");
  xfree(fileName,
	"initLookup: database file");

  /* schedule cron jobs to sync lookup files */
  tmp = rootLevel;
  i = 5;
  while (tmp != NULL) {
    addCronJob((CronJob)&cronSyncLookup,
	       i * cronMINUTES + i * cronSECONDS,
	       i * cronMINUTES,
	       tmp);
    tmp = tmp->next;
    i+=3;
  }
}

/**
 * Shutdown the lookup module.
 **/
void doneLookup() {
  CollisionLevel * tmp;
  int i;

  tmp = rootLevel;
  i = 5;
  while (tmp != NULL) {
    delCronJob((CronJob)&cronSyncLookup,
	       i * cronMINUTES,
	       tmp);
    tmp = tmp->next;
    i+=3;
  }
  tmp = rootLevel;
  while (rootLevel != NULL) {
    tmp = rootLevel;
    rootLevel = rootLevel->next;
    MUTEX_DESTROY(&tmp->lock);
    close(tmp->handle);
    xfree(tmp,
	  "doneLookup, database.X colision level");
  }
}

/**
 * Obtain the offset in the file for the given
 * query.
 **/
static size_t getOffset(CollisionLevel * level,
			HashCode160 * query) {
#if VERY_VERBOSE_LOOKUP
  HexName hex;
  IFLOG(LOG_DEBUG,
	hash2hex(query,
		 &hex));
  LOG(LOG_DEBUG,
      "computing index at level %d mask %x for hash %s\n",
      level->offset, level->mask, 
      &hex);
#endif
  return (((int*)query)[level->offset] & level->mask)
    * sizeof(ContentIndex);
}

/**
 * Delete an entry from the cache. 
 *
 * @param entry the entry to delete
 * @param level the level at which the entry is located,
 *        pass NULL if the level is not known or the entry
 *        is new.
 * @return OK on success, SYSERR on error (errors are not
 *        found or not implemented)
 **/
int delEntry(ContentIndex * entry,
	     CollisionLevel * level) {
  HashCode160 tripleHash;
  HashCode160 * slotQuery;
  int j;
  size_t off;
  ContentIndex cmpEntry;
   
  if (level == NULL)
    level = rootLevel;
  switch (ntohs(entry->type)) {
  case LOOKUP_TYPE_3HASH:
    hash(&entry->hash,
	 sizeof(HashCode160),
	 &tripleHash);
    slotQuery = &tripleHash;
    break;
  case LOOKUP_TYPE_CHK:
  case LOOKUP_TYPE_SUPER:
    slotQuery = &entry->hash;
    break;
  case LOOKUP_TYPE_FREE:
  case LOOKUP_TYPE_DELETED:
    LOG(LOG_WARNING,
	"WARNING: attempt to delete an entry of type free or deleted\n");
    return SYSERR;
  default:
    LOG(LOG_WARNING,
	"WARNING: attempt to delete an entry of unknown type %d\n",
	ntohs(entry->type));    
    return SYSERR;
  }
  while (level != NULL) {
    MUTEX_LOCK(&level->lock);
    off = getOffset(level, slotQuery);
    lseek(level->handle, 
	  off, 
	  SEEK_SET);
    for (j=0;j<LOOKUP_CORUN;j++) {
      read(level->handle, 
	   &cmpEntry,
	   sizeof(ContentIndex));
      if (equalsHashCode160(&entry->hash,
			    &cmpEntry.hash)) {      
	lseek(level->handle, 
	      off, 
	      SEEK_SET);
	memset(entry, 
	       0, 
	       sizeof(ContentIndex));
	entry->type = htons(LOOKUP_TYPE_DELETED);
	write(level->handle, 
	      entry, 
	      sizeof(ContentIndex));
	MUTEX_UNLOCK(&level->lock);
	return OK;
      }    
      off += sizeof(ContentIndex);
    }
    MUTEX_UNLOCK(&level->lock);
    level = level->next;
  } /* while level != NULL */
  return SYSERR;
}

/**
 * Lookup the given Hash in the lookup database structure
 * (load blocks as needed). 
 *
 * @param query the query to look for
 * @param result where to write the lookup information on
 *        success
 * @param level is set to the colevel where the lookup was
 *        successful. This can be used to speed up an update.
 *        Set *level initially to NULL to search the entire
 *        database.
 * @returns OK if found, SYSERR if not
 **/
int findEntry(HashCode160 * query,
	      ContentIndex * result,
	      CollisionLevel ** level) {
  HexName hex;
  HashCode160 tripleHash;
  HashCode160 * slotQuery;
  int j;
  int emptySlot;
  int size;
  
  if (*level == NULL)
    *level = rootLevel;
  emptySlot = NO;
  while (*level != NULL) {
    MUTEX_LOCK(&(*level)->lock);
    lseek((*level)->handle, 
	  getOffset(*level, query), 
	  SEEK_SET);
    for (j=0;j<LOOKUP_CORUN;j++) {
      size = read((*level)->handle, 
		  result, 
		  sizeof(ContentIndex));
      if ( (size < sizeof(ContentIndex)) || 
	   (ntohs(result->type) == LOOKUP_TYPE_FREE) ) {
	emptySlot = YES;
	break;
      }
      if (ntohs(result->type) == LOOKUP_TYPE_3HASH) {
	hash(&result->hash,
	     sizeof(HashCode160),
	     &tripleHash);    
	slotQuery = &tripleHash;
      } else
	slotQuery = &result->hash;      
      if (equalsHashCode160(query, 
			    slotQuery)) {
	MUTEX_UNLOCK(&(*level)->lock);
	return OK;
      }
    }
    MUTEX_UNLOCK(&(*level)->lock);
    if (emptySlot == YES) {
     IFLOG(LOG_DEBUG,
	    hash2hex(query,
		     &hex));
      LOG(LOG_DEBUG,
	  "DEBUG: local lookup did not find a result for %s\n",
	  &hex);
      return SYSERR; /* no collision, nothing there */
    }
  } /* while *level != NULL */
  IFLOG(LOG_DEBUG,
	hash2hex(query,
		 &hex));
  LOG(LOG_DEBUG,
      "DEBUG: local lookup did not find a result for %s\n",
      &hex);
  return SYSERR;  
}

/**
 * Call a method for each entry in the lookup table.
 * VERY expensive (will take quite a while to run),
 * so be careful when using this.
 *
 * @param method the method to call, abort early if 
 *        method returns SYSERR
 * @param data last argument to the callback method
 * @return the number of entries in the database
 **/
int forEachLookupEntry(ContentIndexCallback method,
		       void * data) {
  CollisionLevel * current;
  unsigned int pos;
  ContentIndex entry;
  unsigned int count;
  
  count = 0;
  current = rootLevel;
  while (current != NULL) {
    pos = 0;
    while (pos < current->mask) {
      MUTEX_LOCK(&current->lock);  
      /* we must do an lseek since another
	 thread may do concurrent accesses
	 while we're in the callback method! */
      lseek(current->handle, 
	    pos * sizeof(ContentIndex), 
	    SEEK_SET);
      if (sizeof(ContentIndex) 
	  != read(current->handle,
		  &entry,
		  sizeof(ContentIndex))) {
	MUTEX_UNLOCK(&current->lock);  
 	break; /* end of file */
      }
      MUTEX_UNLOCK(&current->lock);  
      switch (ntohs(entry.type)) {
      case LOOKUP_TYPE_FREE:
      case LOOKUP_TYPE_DELETED:
	break; /* nothing to do */
      case LOOKUP_TYPE_CHK:
      case LOOKUP_TYPE_3HASH:
      case LOOKUP_TYPE_SUPER:
	count++;
	if (SYSERR == method(&entry, 
			     current,
			     pos * sizeof(ContentIndex),
			     data)) {
	  return count;
	}
	break;
      default:
	LOG(LOG_WARNING,
	    "WARNING: entry of unknown type %d in lookup database!\n",
	    ntohs(entry.type));
      }
      pos++;
    }
    current = current->next;
  }
  return count;
}

/**
 * Create a lookup database Iterator.
 **/
LookupDatabaseIterator * createLDI() {
  LookupDatabaseIterator * ldi;

  ldi = xmalloc(sizeof(LookupDatabaseIterator),
		"creating LDI");
  ldi->count = 0;
  ldi->current = rootLevel;
  ldi->pos = 0;
  return ldi;
}

/**
 * Destroy an LDI.
 **/
void destroyLDI(LookupDatabaseIterator * ldi) {
  xfree(ldi,
	"destroying ldi");
}

/**
 * Is there a next element in the database?
 * Always call before next!
 * @return YES if there is, NO if not.
 **/
int ldiHasNext(LookupDatabaseIterator * ldi) {
  while (ldi->current != NULL) {
    MUTEX_LOCK(&ldi->current->lock);  
    lseek(ldi->current->handle, 
	  ldi->pos * sizeof(ContentIndex), 
	  SEEK_SET);
    while (ldi->pos < ldi->current->mask) {
      if (sizeof(ContentIndex) 
	  != read(ldi->current->handle,
		  &ldi->entry,
		  sizeof(ContentIndex))) {
 	break; /* end of file */
      }
      switch (ntohs(ldi->entry.type)) {
      case LOOKUP_TYPE_FREE:
      case LOOKUP_TYPE_DELETED:
	break; /* nothing to do */
      case LOOKUP_TYPE_CHK:
      case LOOKUP_TYPE_3HASH:
      case LOOKUP_TYPE_SUPER:	
	ldi->count++;
	MUTEX_UNLOCK(&ldi->current->lock);  
 	return YES;
      default:
	LOG(LOG_WARNING,
	    "WARNING: entry of unknown type %d in lookup database!\n",
	    ntohs(ldi->entry.type));
      }
      ldi->pos++;
    }
    MUTEX_UNLOCK(&ldi->current->lock);  
    ldi->current = ldi->current->next;
    ldi->pos = 0;
  }
  return NO;
}
  
/**
 * Call the callback method on the next element.
 **/
void ldiGetNext(LookupDatabaseIterator * ldi,
		ContentIndex * entry,
		CollisionLevel ** level) {
  memcpy(entry,
	 &ldi->entry,
	 sizeof(ContentIndex));
  *level = ldi->current;
  ldi->pos++;
}

/**
 * Write the given entry to the database.
 * @param level the level to write at (lock must be held and
 *        will be released!)
 * @param pos the position to write at
 * @param entry the entry to write
 * @return OK on success, SYSERR on error
 **/
static int writeEntry(CollisionLevel * level,
		      size_t pos,
		      ContentIndex * entry) {  
  lseek(level->handle,
	pos,
	SEEK_SET);
  if (sizeof(ContentIndex) == write(level->handle,
				    entry,
				    sizeof(ContentIndex))) {
    MUTEX_UNLOCK(&level->lock);
    return OK;
  } else {
    MUTEX_UNLOCK(&level->lock);
    return SYSERR;
  }
}

/**
 * Add an entry to the cache. An existing entry with the same
 * hash will be overwritten.
 *
 * @param entry the entry to add
 * @param level the level at which the entry is located,
 *        pass NULL if the level is not known or the entry
 *        is new.
 * @return OK on success, SYSERR on error (typical errors
 *   are that the database is full and that we had collisions
 *   throughout all levels of the DB; this should happen if
 *   we index more than CACHE_HASH_TABLE_SIZE blocks of data)
 **/
int addEntry(ContentIndex * entry,
	     CollisionLevel * level) {
  HashCode160 tripleHash;
  HashCode160 * slotQuery;
  int size;
  ContentIndex onDisk;
  int hasFreeBlock;
  int hasDeletedBlock;
  int j;
  size_t pos;
  size_t freePos = 0;
  size_t delPos = 0;

  if (level == NULL)
    level = rootLevel;
  if (ntohs(entry->type) == LOOKUP_TYPE_3HASH) {
    hash(&entry->hash,
	 sizeof(HashCode160),
	 &tripleHash);
    slotQuery = &tripleHash;
  } else 
    slotQuery = &entry->hash;

  while (level != NULL) {
    MUTEX_LOCK(&level->lock);
    pos = getOffset(level, slotQuery);
    lseek(level->handle, 
	  pos,
	  SEEK_SET);
    hasFreeBlock = NO;
    hasDeletedBlock = NO;
    for (j=0;j<LOOKUP_CORUN;j++) {    
      size = read(level->handle, 
		  &onDisk, 
		  sizeof(ContentIndex));
      if ( (size < sizeof(ContentIndex)) ||
	   (ntohs(onDisk.type) == LOOKUP_TYPE_FREE) ) {
	hasFreeBlock = YES;
	freePos = pos;
	break;
      }
      if (ntohs(onDisk.type) == LOOKUP_TYPE_DELETED) {
	hasDeletedBlock = YES;
	delPos = pos;
      }
      if (equalsHashCode160(&onDisk.hash, 
			      &entry->hash) ) 
	return writeEntry(level, pos, entry);
      pos += sizeof(ContentIndex);
    } /* for CORUN */
    if (hasFreeBlock == YES) {
      /* there was a free block, which means we can never
	 have spilled an exact match to the next level;
	 write the entry into the free slot (or into a deleted
	 slot if available) */
      if (hasDeletedBlock == YES)
	freePos = delPos;
      return writeEntry(level, freePos, entry);
    } /* end if hasFreeBlock */
    if (hasDeletedBlock == YES) {
      /* There was no free slot, but there is a deleted slot
	 where we could fit. But we first must check if an
	 equal entry was not spilled to the next level. */
      if (level->next != NULL) {
	CollisionLevel * try;

	try = level->next;
	if (OK == findEntry(slotQuery,
			    &onDisk,
			    &try) ) {
	  /* ok, we had a spill, add the block on that level */
	  MUTEX_UNLOCK(&level->lock);   
	  return addEntry(entry, try);
	}
      } /* end if possibly on next level */
      /* There was no free slot, but also no spill onto the
	 next level. But we had a deleted slot which we can
	 now fill... */
      return writeEntry(level, delPos, entry);
    }
    /* nothing free on this level and nothing deleted
       on this level, move on to the next level */
    MUTEX_UNLOCK(&level->lock);   
    level = level->next;
  }
  return SYSERR; /* all collision levels full! */
}


/* end of lookup.c */
